#include <string.h>
#include <stdbool.h>
#include <stdint.h>

#include "keyboard.h"
#include "multiboot.h"
#include "kprint.h"
#include "cmos.h"
#include "monalisa.h"

#define CGA_SCREEN					((volatile char*)0xb8000)
#define SCRUB_START_ADDR			(3 * 1024 * 1024)

typedef void (*handleMemoryRegion_t)(void *aAddress, uint32_t aLength, void *aParameters);

static void memSet(void *aBaseAddr, uint8_t aChar, uint32_t aLength) {
	uint8_t *base = (uint8_t*)aBaseAddr;
	for (uint32_t i = 0; i < aLength; i++) {
		base[i] = aChar;
	}
}

static void addrFill(void *aBaseAddr, uint32_t aLength) {
	if (aLength < 8) {
		return;
	}

	uint64_t *base = (uint64_t*)aBaseAddr;
	for (uint32_t i = 0; i < aLength / 8; i++) {
		base[i] = (uint32_t)(&base[i]);
	}
}

void patternSet(void *aBaseAddr, const char *aString) {	
	uint8_t *base = (uint8_t*)aBaseAddr;
	while (*aString) {
		*base = *aString;
		aString++;
		base++;
	}
	*base = 0xc0;
	base++;
	*base = 0xff;
	base++;
	*base = 0xee;
}

static void clrScr(void) {
	for (uint32_t i = 0; i < (2 * 80 * 25); i++) {
		CGA_SCREEN[i] = 0;
	}
}

void dumpMemoryMap(multiboot_info_t *multiboot_info) {
	kprintf("Map size %x (%d entries of %d bytes)\n", multiboot_info->mmap_addr, multiboot_info->mmap_length, sizeof(multiboot_memory_map_t));

	multiboot_memory_map_t *maps = (multiboot_memory_map_t*)multiboot_info->mmap_addr;
	for (unsigned int i = 0; i < (multiboot_info->mmap_length / sizeof(multiboot_memory_map_t)); i++) {
		if (maps[i].type == MULTIBOOT_MEMORY_AVAILABLE) {
			kprintf("Entry %d: %c %x len %x %x\n", i, (maps[i].type == MULTIBOOT_MEMORY_AVAILABLE) ? 'A' : 'X', (uint32_t)(maps[i].addr), (uint32_t)(maps[i].len), (uint32_t)(maps[i].addr + maps[i].len));
		}
	}

}

static void memsetMemoryHandler(void *aAddress, uint32_t aLength, void *aParameters) {
	uint8_t *setTo = (uint8_t*)aParameters;
	memSet(aAddress, *setTo, aLength);
}

static void addrfillMemoryHandler(void *aAddress, uint32_t aLength, void *aParameters) {
	(void)aParameters;
	addrFill(aAddress,  aLength);
}

static uint64_t prngNext(void) {
	static uint64_t x = 88172645463325252ull;
	x ^= x << 13;
	x ^= x >> 7;
	x ^= x << 17;
	return x;
}

static void prngfillMemoryHandler(void *aAddress, uint32_t aLength, void *aParameters) {
	(void)aParameters;
	uint64_t *memAddr = (uint64_t*)aAddress;
	for (uint32_t i = 0; i < aLength / 8; i++) {
		*memAddr = prngNext();
		memAddr++;
	}
}

#define PERIOD_BYTES(i, threshold, period)	(((i) % ((period) * 1024)) < (threshold))
#define PERIOD_KB(i, threshold, period)		(((i) % ((period) * 1024)) < ((threshold) * 1024))

static void detfillMemoryHandler(void *aAddress, uint32_t aLength, void *aParameters) {
	(void)aParameters;
	uint8_t *memAddr = (uint8_t*)aAddress;

	uint8_t charPtrn = 0;
	for (uint32_t i = 0; i < aLength; i++) {
		if (PERIOD_KB(i, 512, 1024)) {
			/* Mona Lisa */
			uint32_t offset = i % (1024 * 1024);
			if (offset < MONALISA_SIZE) {
				*memAddr = monalisa[offset];
			} else {
				*memAddr = 0;
			}
		} else {
			if (PERIOD_KB(i, 64, 128)) {
				*memAddr = 0;
			} else if (PERIOD_KB(i, 96, 128)) {
				*memAddr = charPtrn++;
			} else  {
				*memAddr = 0xff;
			}
		}
		memAddr++;
	}
}

static void scrubMemory(multiboot_info_t *multiboot_info, handleMemoryRegion_t handler, void *aHandlerParameters) {
	multiboot_memory_map_t *maps = (multiboot_memory_map_t*)multiboot_info->mmap_addr;
	for (unsigned int i = 0; i < (multiboot_info->mmap_length / sizeof(multiboot_memory_map_t)); i++) {
		if (maps[i].type != MULTIBOOT_MEMORY_AVAILABLE) {
			continue;
		}

		uint64_t startAddr = maps[i].addr;
		uint64_t endAddr = maps[i].addr + maps[i].len;
		if (startAddr < SCRUB_START_ADDR) {
			startAddr = SCRUB_START_ADDR;
		}

		if (startAddr >= endAddr) {
			continue;
		}

		if (startAddr != (startAddr & 0xffffffff)) {
			continue;
		}
		
		if (endAddr != (endAddr & 0xffffffff)) {
			continue;
		}

		uint32_t startOffset = startAddr & 0xffffffff;
		uint32_t length = endAddr - startAddr;
		kprintf("Apply %x len %x\n", startOffset, length);
		handler((void*)startOffset, length, aHandlerParameters);
	}
	kprintf("MEAMORYZ VOLLGEBAUERT! HAVE FUN!\n");
}

static void dumpCMOS(void) {
	uint8_t data[128];
	readCMOS(data);
	for (int i = 0; i < 128; i += 8) {
		kprintf("%02x %02x %02x %02x %02x %02x %02x %02x\n", data[i + 0], data[i + 1], data[i + 2], data[i + 3], data[i + 4], data[i + 5], data[i + 6], data[i + 7]);
		if ((i % 8) == 7) {
			kprintf("\n");
		}
	}
}

static void showMenu(multiboot_info_t *multiboot_info) {
	kprintf("1: Fill memory 0x00\n");
	kprintf("2: Fill memory 0xaa\n");
	kprintf("3: Fill memory with its 8-byte address\n");
//	kprintf("4: Fill memory with PRNG pattern\n");
	kprintf("5: Fill memory with deterministic pattern\n");
//	kprintf("5: Dump CMOS\n");
//	kprintf("9: Dump memory map\n");
	kprintf("0: Restart system\n");
	kprintf("Make your choice!\n");
	
	char key = waitKey();
	clrScr();
	if (key == '1') {
		uint8_t data = 0x00;
		scrubMemory(multiboot_info, memsetMemoryHandler, &data);
	} else if (key == '2') {
		uint8_t data = 0xaa;
		scrubMemory(multiboot_info, memsetMemoryHandler, &data);
	} else if (key == '3') {
		scrubMemory(multiboot_info, addrfillMemoryHandler, NULL);
//	} else if (key == '4') {
//		scrubMemory(multiboot_info, prngfillMemoryHandler, NULL);
	} else if (key == '5') {
		scrubMemory(multiboot_info, detfillMemoryHandler, NULL);
//	} else if (key == '5') {
//		dumpCMOS();
//	} else if (key == '9') {
//		dumpMemoryMap(multiboot_info);
	} else if (key == '0') {
		rebootSystem();
	}
}


/************************************************************************
 * main function, will be called from startup code, must be provided with
 * Multiboot data structures
 ***********************************************************************/
// called from start.S
void kmain(unsigned multiboot_magic, multiboot_info_t *multiboot_info) {
	/* Clear screen */
	clrScr();

	/* Init keyboard */
	initKeyboard();

	/* check whether we have magic and a valid memory map */
	if ((multiboot_magic != MULTIBOOT_BOOTLOADER_MAGIC) || ((multiboot_info->flags | MULTIBOOT_INFO_MEM_MAP) == 0)) {
		/* ... if not we print F on screen and fail */
		CGA_SCREEN[0] = 'F';
		CGA_SCREEN[1] = 0x1e;
		// return to start.S
		return;
	}

	kprintf("Ready.\n");
	while (true) {
		showMenu(multiboot_info);
	}

	// return to start.S
	return;
}


